--- old/usr/src/cmd/ls/ls.c	Mon Jun  5 10:46:46 2017
+++ new/usr/src/cmd/ls/ls.c	Mon Jun  5 10:46:46 2017
@@ -232,8 +232,10 @@
 static int	maxfils = 0;	/* number of flist/lbuf entries allocated */
 static int	maxn = 0;	/* number of flist entries with lbufs asigned */
 static int	quantn = 64;	/* allocation growth quantum */
+static size_t   hlbfsz = 1;
 
 static struct lbuf	*nxtlbf;	/* ptr to next lbuf to be assigned */
+static struct lbuf      **hlbf;         /* lbuf bookkeeping */
 static struct lbuf	**flist;	/* ptr to list of lbuf pointers */
 static struct lbuf	*gstat(char *, int, struct ditem *);
 static char		*getname(uid_t);
@@ -249,6 +251,7 @@
 static void		pem(struct lbuf **, struct lbuf **, int);
 static void		pdirectory(char *, int, int, int, struct ditem *);
 static struct cachenode *findincache(struct cachenode **, long);
+static void             freecachenodes(void);
 static void		csi_pprintf(unsigned char *);
 static void		pprintf(char *, char *);
 static int		compar(struct lbuf **pp1, struct lbuf **pp2);
@@ -413,6 +416,7 @@
 	int		amino = 0;
 	int		opterr = 0;
 	int		option_index = 0;
+	char            *told = NULL;
 	struct lbuf	*ep;
 	struct lbuf	lb;
 	struct ditem	*myinfo = NULL;
@@ -634,7 +638,7 @@
 					continue;
 				}
 				if (optarg[0] == '+') {
-					char	*told, *tnew;
+					char	*tnew;
 					char	*p;
 					size_t	timelen = strlen(optarg);
 
@@ -650,7 +654,7 @@
 					 * a space.
 					 */
 					timelen += 2 + 3;
-					told = malloc(timelen);
+					told = realloc(told, timelen);
 					if (told == NULL) {
 						perror("ls");
 						exit(2);
@@ -664,10 +668,11 @@
 
 					if (p != NULL) {
 						size_t tnew_len;
+						size_t told_len =strlen(told);
 
-						tnew = told + strlen(told) + 1;
+						tnew = told + told_len + 1;
 						tnew_len = timelen -
-						    strlen(told) - 1;
+						    told_len - 1;
 
 						tnew[0] = ' ';
 						(void) strlcat(tnew, p,
@@ -1022,6 +1027,11 @@
 		perror("ls");
 		exit(2);
 	}
+	if ((hlbf = malloc(sizeof(*hlbf))) == NULL) {
+		perror("ls");
+		exit(2);
+	}
+	hlbf[0] = nxtlbf;
 	if ((amino = (argc-optind)) == 0) {
 					/*
 					 * case when no names are given
@@ -1129,6 +1139,14 @@
 		}
 	}
 
+	for (i = 0; i < hlbfsz; i ++)
+		free(hlbf[i]);
+
+	free(told);
+	free(hlbf);
+	free(flist);
+	freecachenodes();
+
 	return (err);
 }
 
@@ -1797,6 +1815,12 @@
 			nomocore = 1;
 			return (NULL);
 		}
+		if ((hlbf = realloc(hlbf, sizeof(*hlbf) * (hlbfsz + 1))) == NULL) {
+			perror("ls");
+			nomocore = 1;
+			return (NULL);
+		}
+		hlbf[hlbfsz++] = nxtlbf;
 	}
 
 	/*
@@ -2221,6 +2245,36 @@
 	return (c);
 }
 
+void
+freecachenode(struct cachenode *node)
+{
+	struct cachenode *current = node;
+	if (current != NULL) {
+		struct cachenode *grt = NULL;
+		struct cachenode *lss = NULL;
+
+		if (current->grtrchild != NULL) {
+			grt = current->grtrchild;
+			freecachenode(grt);
+		}
+		if (current->lesschild != NULL) {
+			lss = current->lesschild;
+			freecachenode(lss);
+		}
+
+		free(current);
+		current = NULL;
+	}
+}
+
+void
+freecachenodes(void)
+{
+	freecachenode(groups);
+	freecachenode(names);
+}
+
+
 /*
  * get name from cache, or passwd file for a given uid;
  * lastuid is set to uid.
@@ -2623,7 +2677,7 @@
 	size_t		len;
 
 	if (nvpair_value_uint64_array(pair, &value, &nelem) == 0) {
-		if (*value != NULL) {
+		if (value != NULL) {
 			len = strlen(name);
 			i = 0;
 			while (rep->extm[i].stm != 0 && i < sacnt)
